# encoding: UTF-8

module Dyndoc

  @@tmpl_mngr=nil

  ##need to point to the current TemplateManager
  def Dyndoc.tmpl_mngr
    @@tmpl_mngr
  end

  def Dyndoc.tmpl_mngr=(tmpl_mngr)
    @@tmpl_mngr = tmpl_mngr
  end

  ## Make more readable the creation of newBlck
  def Dyndoc.cache(cache_={})
    return Dyndoc.tmpl_mngr.vars[Dyndoc.tmpl_mngr.blckName[-1]] if cache_.empty?
    Dyndoc.tmpl_mngr.vars[Dyndoc.tmpl_mngr.blckName[-1]]=cache_
  end

  def Dyndoc.extract_var(id)
    Dyndoc.tmpl_mngr.vars.extract(id)
  end

  module Ruby

  class TemplateManager

    attr_accessor :global, :filterGlobal, :filter, :vars, :libs , :tmpl , :blocks, :calls, :args, :envirs, :fmt, :Fmt, :fmtContainer, :fmtOutput, :dyndocMode ,:echo, :filename, :tags, :strip
    attr_accessor :tmpl_cfg, :cfg_tmpl, :cfg, :doc
    attr_accessor :rbEnvir, :rbEnvirs,  :rEnvir
    attr_accessor :rbBlock
    attr_reader :scan, :blckName
    ### attr_accessor :mark

    @@interactive=nil

    def TemplateManager.interactive
      @@interactive=Dyndoc.cfg_dyn[:dyndoc_session]==:interactive unless @@interactive
      @@interactive
    end

    # Maybe better located inside server.rb
    def TemplateManager.initR
      ## THIS IS JUST IMPOSSIBLE SINCE RUBY USES HOME TO EXPAND PATH
      ## NEEDS TO FIND AN ALTERNATIVE SOLUTION: R_LIBS_USER
      # For windows! R now change HOME variable by appending "/Documents" at the end. Ruby does not do that!
      if RUBY_PLATFORM =~ /mingw/ and !(File.exist? File.join(ENV["HOME"],"R","win-library")) and (File.exist? File.join(ENV["HOME"],"Documents","R","win-library"))
        ##ENV["HOME"] += "/Documents"
        if ENV["R_LIBS_USER"]
          puts "Warning (dyndoc): R_LIBS_USER already in use by dyndoc. Fix that by calling inside dyndoc .libPaths(\"#{ENV['R_LIBS_USER']}\")"
        end
        ENV["R_LIBS_USER"]=Dir[File.join(ENV["HOME"],"Documents","R","win-library","*")][-1]
        ## RMK: if R_LIBS_USER already exist then the second solution
      end

      first=require "R4rb" #save if it the first initialization!
      Dyndoc.warn "FIRST INIT OF R!!!! => #{first}"
      Array.initR
      TemplateManager.interactive
      #p "client";p Dyndoc.cfg_dyn;p interactive
      R4rb << "rm(list=ls(all=TRUE))" if !first and !@@interactive #remove all initial variables if previous documents session
      R4rb << ".dynStack<-new.env()" #used for R variables used by dyndoc
      RServer.init_envir
      RServer.init_filter
      ## ruby and R init for dynArray stuff
      require "dyndoc/common/dynArray"
      dyndocToolsR = File.join(Dyndoc.cfg_dir[:gem_path],"share","R","dyndocTools.R").gsub('\\','/')
      ## WD is introduced by Msys and ConEmuDir by ConEmu
      dyndocToolsRoot = RUBY_PLATFORM =~ /msys/ ? (ENV['WD'] || ENV['ConEmuDir']).split('\\')[0...-2].join(File::Separator) : ""
      R4rb << "source('"+dyndocToolsRoot+dyndocToolsR+"',chdir=TRUE)"
      if RUBY_PLATFORM =~ /msys/
        dyndocMsys2R = File.join(Dyndoc.cfg_dir[:gem_path],"share","R","dyndocMsys2.R").gsub('\\','/')
        R4rb << "source('"+dyndocToolsRoot+dyndocMsys2R+"',chdir=TRUE)"
      end
      if RUBY_PLATFORM =~ /mingw/
        require File.join(ENV["DYNDOC_HOME"],"bin","msys2","file_tools.rb") if ENV["DYNDOC_HOME"]
      end
      Dyndoc.logger.info "R initialized inside dyndoc!"
    end

    def TemplateManager.initJulia
      Dyndoc.logger.info "initJulia: first"
      first=require "jl4rb" #save if it the first initialization!
      Dyndoc.logger.info "initJulia: require"
      Julia.init
      Dyndoc.logger.info "initJulia: init"
      # init rb4jl stuff
      # since inside ruby, no need Ruby.start and Ruby.stop like in rb4R.
      # sort of equivalent of JLServer.init_filter (but not yet defined)!
      #Julia << "Base.include(@__MODULE__,\"ruby.jl\")"
      Julia << "Base.include(@__MODULE__,\""+File.join(Dyndoc.cfg_dir[:gem_path],"share","julia","ruby.jl")+ "\")"
      Dyndoc.logger.info "initJulia: ruby.jl"
      Julia << "Base.include(@__MODULE__,\""+File.join(Dyndoc.cfg_dir[:gem_path],"share","julia","dynArray.jl")+ "\")"
      Dyndoc.logger.info "initJulia: dynArray.jl"
      #-| To debug ruby.jl and dynArray.jl => uncomment below and commnt above
      # Julia << "include(\""+File.expand_path("~/Github/dyndoc/share/julia/ruby.jl")+"\")"
      # Julia << "include(\""+File.expand_path("~/Github/dyndoc/share/julia/dynArray.jl")+"\")"
      Julia << "using Main.Dyndoc"
      Dyndoc.logger.info "initJulia: using Dyndoc"
      Julia << "Dyndoc.Ruby.alive(true)"
      #Julia << "global const _dynArray=DynArray()"
      Dyndoc.logger.info "Julia initialized inside dyndoc!"

    end

    def TemplateManager.attr
      attr={:cmd => @@cmd,:cmdAlias => @@cmdAlias, :argsSep => @@argsSep, :tags => @@tags,:prefix => @@prefix, :tagSearch => @@tagSearch, :tagModifiers => @@tagModifiers}
      attr[:tags_tex],attr[:tex_vars] = @@tags_tex,@@tex_vars if TemplateManager.class_variables.include? "@@tag_tex"
      attr
    end

    def initialize(tmpl_cfg={},with=true)
      # just in case it is not yet initialized!
      (Dyndoc.cfg_dyn[:langs] << :R).uniq! if with==true

#puts "DEBUT INIT TemplateManager"
      @tmpl_cfg=tmpl_cfg
=begin
      @cfg[:part_tag][0]=@cfg[:part_tag][0][1..-1] if !(@cfg[:part_tag].empty?) and (@partTag_add= (@cfg[:part_tag][0][0,1]=="+"))
=end
      ## default system root appended
      unless  Dyndoc.cfg_dyn[:ruby_only]
        TemplateManager.initR if Dyndoc.cfg_dyn[:langs].include? :R
        TemplateManager.initJulia if Dyndoc.cfg_dyn[:langs].include? :jl
      end
      rbenvir_init(binding)
      @rEnvir=["Global"]
      @envirs={}
      @fmtContainer=[]
      @echo=1
      @strip=true
#puts "FIN INIT TemplateManager"
      if Dyndoc.cfg_dyn[:devel_mode]==:test
        puts "DYNDOC SEARCH PATH:"
        puts Dyndoc.get_pathenv(Dyndoc.cfg_dyn[:root_doc]).join(":")
      end
    end

    ## Make more readable the creation of newBlck
    def cache(cache_={})
      return @vars[blckName[-1]] if cache_.empty?
      @vars[blckName[-1]]=cache_
    end

    def extract_var(id)
      @vars.extract(id)
    end

    ## equivalent to :pre_doc for not interactive document
    def require_first
      require_dyndoc_libs("Dyn/Base")
      Dyndoc.cfg_dyn[:require_first].each do |libs|
        require_dyndoc_libs(libs)
      end
    end

    def require_dyndoc_libs(libs)
      libs="{#require]\n"+libs.split("\n").map{|lib| lib.split(",")}.flatten.uniq.join("\n")+"\n[#}\n"
      parse(libs,@filterGlobal)
    end

    ## to set the current tmpl_mngr as the default one accessed via Dyndoc.tmpl_mngr
    ## and used in #R> #rb> block
    def as_default_tmpl_mngr!
      Dyndoc.tmpl_mngr=self unless Dyndoc.tmpl_mngr == self
    end

    def init_doc(doc_cfg)
      @cfg=doc_cfg
      # register format
      @fmt=@cfg[:format_doc].to_s.downcase
      @Fmt=@fmt.capitalize
      ##Dyndoc.warn "@cfg",@cfg
      @fmtOutput=@cfg[:format_output].to_s if @cfg[:format_output]
      @fmtOutput=@fmt if @fmt and ["html","tex","odt","tm"].include? @fmt
      @dyndocMode=:cmdline
      @global={}
      @filterGlobal=FilterManager.global(self) #global filter needed to affect object before parsing the main document!
      @blocks,@libs,@calls,@args,@meths,@def_blck={},{},{},{},[],[]
      CallFilter.init(self,@calls,@args,@meths)
      @filter=nil
      @tags=[]
      @keys={}
      @alias={}
      @savedBlocks={}
      require_first if @@interactive
    end

    def format_output=(format)
      @fmtOutput=format
    end

    def dyndoc_mode=(mode)
      @dyndocMode=mode
    end

    def reinit
      @blocks,@libs={},{}
    end

    ## init output
    def init_path(input)
       ##read in the main doc paths to include!
      Dyndoc.setRootDoc(@cfg[:rootDoc],input.scan(/%%%path\((.*)\)/).flatten.map{|e| e.split(",")}.flatten.map{|e| e.strip.downcase}.join(":"))
      p @cfg[:rootDoc] if @cfg[:cmd]==:cfg
    end

    def init_tags(input)
      ## read the global aliases (config files alias)
      TagManager.global_alias(@alias) # @alias increased
      ## read the local aliases (in the main doc)
      TagManager.local_alias(@alias,input) # @alias increased
      #init @partTag
      @tags=TagManager.init_input_tags(([@fmt]+@cfg[:tag_doc]+@tmpl_cfg[:tag_tmpl]).uniq)
#p @alias
#
Dyndoc.warn "init_tags",@tags
      #To deal later: TagManager.apply_alias(@tags,@alias)
      p [:init_tags, @tags] if @cfg[:cmd]==:cfg
    end

    def init_keys
      @init_keys=KeysManager.init_keys((@cfg[:keys_doc]+@tmpl_cfg[:keys_tmpl]).uniq)
#puts "Manager:init_keys";p @init_keys
    end


    def init_model(input)
      @pre_doc,@post_doc,@pre_model,@post_model=[],[],[],[]
      if @cfg[:model_doc] and Dyndoc.cfg_dir[:tmpl_path][@cfg[:format_doc]]
        @cfg[:model_doc]="Default" if @cfg[:model_doc].downcase=="default"
        model_doc=File.join("Model",Dyndoc.cfg_dir[:tmpl_path][ @cfg[:format_doc]],@cfg[:model_doc])
        @pre_model << File.read(Dyndoc.doc_filename(model_doc+"Pre"))
        @post_model = [File.read(Dyndoc.doc_filename(model_doc+"Post"))]
#p @pre_model
#p @post_model
      end

      ## sort with respect to priority number and filter
      @cfg[:pre_doc] += input.scan(/\[\#(?:plugin|preload)\]([^\[]*)/m).flatten.map{|e| e.strip.split("\n").map{|e2| e2.strip}}.flatten
#p @cfg[:pre_doc]
      input.gsub!(/\[\#(?:plugin|preload)\][^\[]*/m,"")
      @cfg[:pre_doc].sort!.map!{|e| e.scan(/^\d*(.*)/)}.flatten!
      ## sort with respect to priority number and filter
      @cfg[:post_doc] += input.scan(/\[\#(?:postload)\]([^\[]*)/m).flatten.map{|e| e.strip.split("\n").map{|e2| e2.strip}}.flatten
      input.gsub!(/\[\#(?:postload)\][^\[]*/m,"")
      @cfg[:post_doc].sort!.map!{|e| e.scan(/^\d*(.*)/)}.flatten!

#p @cfg[:pre_doc]
      if @cfg[:pre_doc]
        @cfg[:pre_doc].uniq.each{|t|
          @pre_doc << File.read(Dyndoc.doc_filename(t))
        }
      end
#p @pre_doc

      if @cfg[:post_doc]
        @cfg[:post_doc].uniq.each{|t|
          @post_doc << File.read(Dyndoc.doc_filename(t))
        }
      end

    end

    def output_pre_model
      #pre_doc
      out_pre=""
      unless @pre_model.empty?
        pre_model=@pre_model.join("\n") + "\n"
        txt= pre_model
        out_pre << parse(txt)
      end
      unless @pre_doc.empty?
        pre_doc=@pre_doc.join("\n") + "\n"
        txt= pre_doc
        parse(txt,@filterGlobal) ##Style declares objects before compiling the document! Object are created in the global envir!
      end
##p [:out_pre,out_pre]
      return out_pre
    end
=begin
      input = pre_doc.join("\n") + "\n" + input unless pre_doc.empty?
      input = input + "\n" + post_doc.join("\n") unless post_doc.empty?
      return input
=end

    def output_post_model
     #post_doc
      out_post=""
      #puts "output_post_model:post doc";p @post_doc
      unless @post_doc.empty?
        post_doc="\n"+@post_doc.join("\n")
        txt= post_doc
        parse(txt)
      end
      unless @post_model.empty?
        post_model="\n"+@post_model.join("\n")
        txt= post_model
        out_post << parse(txt)
      end
#p [:out_post,out_post]
      return out_post
    end


    ## TO REMOVE????
    def init_dtag(tmpl)
      dtag=tmpl.scan(/(?:#{@cfg[:dtags].map{|e| "_#{e}_" }.join("|")})/)[0]
#p dtag
      return @dtag=(dtag ? dtag[1...-1].to_sym : @cfg[:dtag])
    end

    def prepare_output(txt)
      ## parse the content txt
      out=output_pre_model
#p out
#p Dyndoc.dyn_block[:first]
      ## dyn_block[:first] called in preloaded libraries!
      out << parse(Dyndoc.dyn_block[:first]) if Dyndoc.dyn_block[:first]
#puts "prepare_output";p out
      out << parse(txt)
#puts "prepare_output";p out
      ## dyn_block[:last] (for example, used to record some information)
      out << parse(Dyndoc.dyn_block[:last]) if Dyndoc.dyn_block[:last]
      out << output_post_model
      ## escape=true at the end in order to transform undefined variables in Latex!!!
##puts "prepare output: BEFORE apply";puts out
      out=@filter.apply(out,:post,false,true) if [:tex,:odt,:ttm,:html].include? @cfg[:format_doc]
##p [:prepare_output, out]
      #escape accolade!
      Utils.escape!(out,CHARS_SET_LAST)
##puts "prepare output: AFTER Utils.escape!";puts out
      out=Utils.unprotect_format_blocktext(out)
##p [:prepare_output, out]
      return out
    end

    ##only once!!! Not like prepare_output which is called in {#dyn]pre:
    def prepare_last_output(out)
      #puts "DYNDOC_RAW_TEXT! BEFORE";p out
      #p Utils.dyndoc_raw_text
      Utils.dyndoc_raw_text!(out,:clean=>true)
      #puts "DYNDOC_RAW_TEXT! AFTER";p out
      return out
    end

    def clean_as_is(out)
      out.gsub!(Dyndoc::AS_IS,"")
    end

    def prepare_user_input
      ##puts "input";p @cfg[:input];p Dyndoc.cfg_dyn[:user_input]
      if @cfg[:input]
        @global["_"]={} unless @global["_"]
        @cfg[:input].each{|k,v|
          @global["_"][k.to_s] = {:val=>[v]}
        }
      end
      if Dyndoc.cfg_dyn[:user_input]
        @global["_"]={} unless @global["_"]
        Dyndoc.cfg_dyn[:user_input].each{|k,v|
          @global["_"][k.to_s] = {:val=>[v]}
        }
      end
    end

    ## output the result after parsing txt
    def output(input,fromfile=nil)
      @filename=@tmpl_cfg[:filename_tmpl]
      input=Dyndoc.input_from_file(@filename=input) if fromfile
      input=@tmpl_cfg[:lib_dyn_content]+"[#main][#>]"+input unless @tmpl_cfg[:lib_dyn_content].empty?
      ## fetch the contents of all saved variables! This fasten the compilation!
      Utils.saved_content_fetch_variables_from_file(@filename,self)

#p @filename
      init_path(input)
      init_tags(input)
      init_keys
      init_model(input)
#p input
      ## add the tag document in order to replace the user tag preamble and etc. Introduction of styles because of the odt format and the new convention in general to use styles even in latex.
      #txt="{#document][#content]"+input+"[#}"
      txt=input
#p txt
      prepare_user_input
      ## parse the content txt
      out=prepare_output(txt)
      prepare_last_output(out)
      clean_as_is(out)
##p [:output,out]
      return out
    end

=begin
    ## DONE: put in Document class
    ## writing in final file
    def write(file)
      #@cfg[:append]="" unless @cfg[:append]
      out = output(file,true)
#puts out
      f=File.open(@cfg[:filename_doc],"w")
      f << out
      f.close
    end
=end

    def query_library(libname)
      init_doc({:format_doc=>:tex})
      parse(Dyndoc.input_from_file(libname))
      p @calls.keys.sort
      p @meths.sort
    end

  end

  end
end

# String helpers ######################

class String

  def to_keys
    Dyndoc::Envir.to_keys(self)
  end

  def to_dyn
    if Dyndoc.tmpl_mngr
      return Dyndoc.tmpl_mngr.parse_string(self)
    else
      return self
    end
  end

  def to_var
    if Dyndoc.tmpl_mngr and (res=Dyndoc.tmpl_mngr.vars.extract(self))
      res
    else
      self
    end
  end

  def to_var=(val)
    if Dyndoc.tmpl_mngr
      Dyndoc.tmpl_mngr.vars[self]=val
    end
  end

end

# Symbol helpers ##################################

class Symbol

  def to_keys
    to_s.to_keys
  end

  def to_var
    if Dyndoc.tmpl_mngr and (res=Dyndoc.tmpl_mngr.vars.extract(self.to_s))
      res
    else
      self
    end
  end

  def to_var=(val)
    if Dyndoc.tmpl_mngr
      Dyndoc.tmpl_mngr.vars[self.to_s]=val
    end
  end

end
